#include "hid.h"
#include "ruby.h"

static VALUE cHid;
static VALUE cHidCaps;
static VALUE cHidAttributes;
static VALUE eHidError;
static GUID hidGuid;

static HMODULE hHidDll;
static HMODULE hSetupDll;

static GetGuidProc      getGuid;
static GetAttrProc      getAttr;
static GetPreparsedProc getPreparsed;
static GetCapsProc      getCaps;

static SetupDiGetClassDevsProc             getClassDevs;
static SetupDiEnumDeviceInterfacesProc     enumDeviceInterfaces;
static SetupDiGetDeviceInterfaceDetailProc getDeviceInterfaceDetail;
static SetupDiDestroyDeviceInfoListProc    destroyDeviceInfoList;


void _inner_hid_load_dlls( void );
DWORD _inner_hid_queue_read( hid_t* hid );
DWORD _inner_hid_finish_read( hid_t* hid );
PSP_INTERFACE_DEVICE_DETAIL_DATA _inner_GetHidDeviceDetail( hid_t *hid, int index );
void _inner_hid_open( hid_t* hid );



/**
 * destruction
 */
static void hid_free( hid_t *hid )
{
	free( hid->input_buffer  );
	free( hid->output_buffer );
	
	if ( hid->hDevice     != INVALID_HANDLE_VALUE ) { CloseHandle( hid->hDevice     ); }
	if ( hHidDll          != INVALID_HANDLE_VALUE ) { CloseHandle( hHidDll          ); }
	if ( hSetupDll        != INVALID_HANDLE_VALUE ) { CloseHandle( hSetupDll        ); }
	
	free( hid );
}


/**
 * write data
 */
static VALUE hid_write( VALUE self, VALUE data )
{
	int i;
	hid_t *hid;
	int count;
	VALUE result;
	Data_Get_Struct( self, hid_t, hid );
	
	if ( RSTRING( data )->len > hid->caps.OutputReportByteLength ) {
		count = hid->caps.OutputReportByteLength;
	}
	else
	{
		count = RSTRING(data)->len;
	}
	// Zero clear
	ZeroMemory( hid->output_buffer, hid->caps.OutputReportByteLength );
	// copy data
	memcpy( hid->output_buffer, RSTRING(data)->ptr, count );
	// write data
	if ( WriteFile( hid->hDevice, hid->output_buffer, hid->caps.OutputReportByteLength, &count, &hid->overlapped ) == 0 ) {
		// unexpected error happened
		rb_raise( eHidError, "cannot write data to the device." );
		return -1;
	}
	
	return count;
}

/**
 * read data
 */
static VALUE hid_read( VALUE self )
{
	hid_t *hid;
	int count;
	VALUE result;
	Data_Get_Struct( self, hid_t, hid );
	
	count = _inner_hid_finish_read( hid );
	
	return rb_str_new( hid->input_buffer, count );
}


/**
 * create a HID object, and open a HID device.
 */
static VALUE hid_singleton_new( int argc, VALUE *argv, VALUE self )
{
	VALUE obj, product_id, vendor_id;
	hid_t *hid;
	
	rb_scan_args( argc, argv, "20", &vendor_id, &product_id );
	
	obj = Data_Make_Struct( self, hid_t, NULL, hid_free, hid );
	
	hid->hDevice       = INVALID_HANDLE_VALUE;
	hid->input_buffer  = NULL;
	hid->output_buffer = NULL;
	
	hid->vendor_id     = NUM2INT( vendor_id  );
	hid->product_id    = NUM2INT( product_id );

	_inner_hid_open( hid );
	
	return obj;
}


void Init_hid( void )
{
	_inner_hid_load_dlls( );
	
	getGuid( &hidGuid );
	
	
	cHid = rb_define_class( "Hid", rb_cObject );
	rb_define_singleton_method( cHid, "new", hid_singleton_new, -1 );
	rb_define_method( cHid, "read", hid_read, 0 );
	rb_define_method( cHid, "write", hid_write, 1 );
	
	/*
	cHidCaps = rb_define_class_under( cHid, "Caps", rb_cObject );
	rb_define_singleton_method( cHidCaps, "new", hid_caps_singleton_new, -1 );
	
	cHidAttributes = rb_define_class_under( cHid, "Attributes", rb_cObject );
	rb_define_singleton_method( cHidAttributes, "new", hid_attributes_singleton_new, -1 );
	*/
	
	eHidError = rb_define_class_under( cHid, "Error", rb_eStandardError );
}


//----------------------------------------------------------------------------//
//--                             inner functions                            --//
//----------------------------------------------------------------------------//


/**
 * load dll functions dynamically
 */
void _inner_hid_load_dlls( void )
{
	hHidDll = INVALID_HANDLE_VALUE;
	hSetupDll = INVALID_HANDLE_VALUE;
	// @cite http://www.google.com/codesearch?hl=ja&q=+HIDSDI.H+show:OBg4C5O9BGk:K_0lhEsO1a8:qhVMFkmN3oQ&sa=N&cd=7&ct=rc&cs_p=http://cq.cx/dl/at91sam7sXXX.zip&cs_f=loader/usbdl.cpp#a0
	hHidDll = LoadLibrary( L"hid.dll" );
	if ( hHidDll == INVALID_HANDLE_VALUE )
	{
	  rb_raise( eHidError, "cannot load 'hid.dll'." );
	  return;
	}

	getGuid      = ( GetGuidProc      )GetProcAddress( hHidDll, "HidD_GetHidGuid" );
	getAttr      = ( GetAttrProc      )GetProcAddress( hHidDll, "HidD_GetAttributes" );
	getPreparsed = ( GetPreparsedProc )GetProcAddress( hHidDll, "HidD_GetPreparsedData" );
	getCaps      = ( GetCapsProc      )GetProcAddress( hHidDll, "HidP_GetCaps" );
	
	hSetupDll = LoadLibrary( L"setupapi.dll" );
	if ( hSetupDll == INVALID_HANDLE_VALUE )
	{
	  rb_raise( eHidError, "cannot load 'setupapi.dll'." );
	  return;
	}
	getClassDevs             = ( SetupDiGetClassDevsProc             )GetProcAddress( hSetupDll, "SetupDiGetClassDevsW" );
	enumDeviceInterfaces     = ( SetupDiEnumDeviceInterfacesProc     )GetProcAddress( hSetupDll, "SetupDiEnumDeviceInterfaces" );
	getDeviceInterfaceDetail = ( SetupDiGetDeviceInterfaceDetailProc )GetProcAddress( hSetupDll, "SetupDiGetDeviceInterfaceDetailW" );
	destroyDeviceInfoList    = ( SetupDiDestroyDeviceInfoListProc    )GetProcAddress( hSetupDll, "SetupDiDestroyDeviceInfoList" );
}



/**
 * asynchronous read reservation
 */
static DWORD _inner_hid_queue_read( hid_t* hid )
{
	DWORD count;
	ReadFile( hid->hDevice, hid->input_buffer, hid->caps.InputReportByteLength, &count, &hid->overlapped );
	
	return count;
}


/**
 * asynchronous read
 */
static DWORD _inner_hid_finish_read( hid_t* hid )
{
	DWORD count;
	
	if ( GetOverlappedResult( hid->hDevice, &hid->overlapped, &count, FALSE ) != 0 )
	{
		// data is ready
		_inner_hid_queue_read( hid );
		
		return count;
	}
	else
	{
		// data is not ready
		DWORD error_code = GetLastError( );
		if ( error_code == ERROR_IO_PENDING || error_code == ERROR_IO_INCOMPLETE )
		{
			return 0;
		}
		// unexpected error happened
		rb_raise( eHidError, "cannot read data from the device." );
	}
	return -1;
}


/**
 * get a HID device detail by calling Windows API
 */
static PSP_INTERFACE_DEVICE_DETAIL_DATA _inner_GetHidDeviceDetail( hid_t *hid, int index )
{
	PSP_INTERFACE_DEVICE_DETAIL_DATA DeviceDetail;
	DWORD size;
	/*
	fprintf( stderr, "device_info: %08x\n", (unsigned int)(hid->device_info) );
	fprintf( stderr, "device_interface: %08x\n", (unsigned int)(hid->p_device_interface) );
	*/
	if ( !enumDeviceInterfaces( hid->device_info, NULL, &hidGuid, index, hid->p_device_interface ) )
	{	
		return NULL;
	}

	if ( getDeviceInterfaceDetail( hid->device_info, hid->p_device_interface, NULL, 0, &size, 0 ) )
	{
		if ( GetLastError( ) != ERROR_INSUFFICIENT_BUFFER )
		{
			return NULL;
		}
	}
	
	DeviceDetail = ( PSP_INTERFACE_DEVICE_DETAIL_DATA )ALLOC_N( char, size );
	DeviceDetail->cbSize = sizeof( SP_INTERFACE_DEVICE_DETAIL_DATA );
	
	if ( !getDeviceInterfaceDetail( hid->device_info, hid->p_device_interface, DeviceDetail, size, &size, NULL ) ) 
	{
		fprintf( stderr, "\n\nerror code: %d\n\n", GetLastError( ) );
		return NULL;
	}
	
	return DeviceDetail;
}


/**
 * open a HID device which specified with VendorID & ProductID
 * 
 * @todo extend to multiple device by indexing
 */
static void _inner_hid_open( hid_t* hid )
{
	int index;
	HANDLE handle;
	PSP_INTERFACE_DEVICE_DETAIL_DATA DeviceDetail;
	
	hid->device_info = getClassDevs( &hidGuid, NULL, NULL, ( DIGCF_PRESENT | DIGCF_DEVICEINTERFACE ) );

	hid->p_device_interface = ( PSP_DEVICE_INTERFACE_DATA )ALLOC( SP_DEVICE_INTERFACE_DATA );
	hid->p_device_interface->cbSize = sizeof( SP_DEVICE_INTERFACE_DATA );

	for( index = 0; ; index++ )
	{
		//fprintf( stderr, "search: %d\n", index );
		DeviceDetail = _inner_GetHidDeviceDetail( hid, index );
		//	TIĂ܂
		if ( DeviceDetail == NULL )
		{
		  rb_raise( eHidError, "cannot find expected device." );
		  break;
		}
		
		handle = CreateFile( DeviceDetail->DevicePath, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL );
		if ( handle == INVALID_HANDLE_VALUE )
		{
			//fprintf( stderr, "  cannot open device, %4x\n", GetLastError( ) );
			continue;
		}
		
		if ( !getAttr( handle, &hid->attributes ) )
		{
			fprintf( stderr, "  cannot get attributes: %4x\n", GetLastError( ) );
			continue;
		}

		if ( !getPreparsed( handle, &hid->preparsed_data ) )
		{
		  rb_raise( eHidError, "cannot get PreparsedData." );
		  break;
		}
		if ( getCaps( hid->preparsed_data, &hid->caps ) != HIDP_STATUS_SUCCESS )
		{
		  rb_raise( eHidError, "cannot get Caps." );
		  break;
		}
		CloseHandle( handle );
		/*
			fprintf( stderr, "  expected: vid: %02x, pid: %02x\n", hid->vendor_id, hid->product_id );
			fprintf( stderr, "  actually: vid: %02x, pid: %02x\n", hid->attributes.VendorID, hid->attributes.ProductID );
			fprintf( stderr, "    result: vid: %02d, pid: %02d\n", hid->attributes.VendorID != hid->vendor_id, hid->attributes.ProductID != hid->product_id );
		*/
		if ( hid->attributes.VendorID != hid->vendor_id || hid->attributes.ProductID != hid->product_id )
		{
			continue;
		}
		
		//  񓯊 IO ɂ̂ŁAFILE_FLAG_OVERLAPPED tĊJB
		handle = CreateFile( DeviceDetail->DevicePath, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL );
		if ( handle == INVALID_HANDLE_VALUE )
		{
		  rb_raise( eHidError, "cannot open device." );
		  break;
		}
		//free( DeviceDetail );
		DeviceDetail = NULL;
		
		hid->hDevice = handle;
		
		hid->input_buffer  = ALLOC_N( char, hid->caps.InputReportByteLength  + 1 );
		hid->output_buffer = ALLOC_N( char, hid->caps.OutputReportByteLength + 1 );
		/*
		fprintf( stderr, "\n" );
		fprintf( stderr, "Device Caps\n" );
		fprintf( stderr, "------------------------------------\n" );
		fprintf( stderr, "Reserved    : %d\n", hid->caps.Reserved  );
		fprintf( stderr, "Usage       : %d\n", hid->caps.Usage     );
		fprintf( stderr, "UsagePage   : %d\n", hid->caps.UsagePage );
		fprintf( stderr, "\n" );
		fprintf( stderr, "Input Bytes             : %d\n", hid->caps.InputReportByteLength    );
		fprintf( stderr, "NumberInputButtonCaps   : %d\n", hid->caps.NumberInputButtonCaps    );
		fprintf( stderr, "NumberInputDataIndices  : %d\n", hid->caps.NumberInputDataIndices   );
		fprintf( stderr, "NumberInputValueCaps    : %d\n", hid->caps.NumberInputValueCaps     );
		fprintf( stderr, "\n" );
		fprintf( stderr, "Feature Bytes           : %d\n", hid->caps.FeatureReportByteLength  );
		fprintf( stderr, "NumberFeatureButtonCaps : %d\n", hid->caps.NumberFeatureButtonCaps  );
		fprintf( stderr, "NumberFeatureDataIndices: %d\n", hid->caps.NumberFeatureDataIndices );
		fprintf( stderr, "NumberFeatureValueCaps  : %d\n", hid->caps.NumberFeatureValueCaps   );
		fprintf( stderr, "\n" );
		fprintf( stderr, "Output Bytes            : %d\n", hid->caps.OutputReportByteLength   );
		fprintf( stderr, "NumberOutputButtonCaps  : %d\n", hid->caps.NumberOutputButtonCaps   );
		fprintf( stderr, "NumberOutputDataIndices : %d\n", hid->caps.NumberOutputDataIndices  );
		fprintf( stderr, "NumberOutputValueCaps   : %d\n", hid->caps.NumberOutputValueCaps    );
		fprintf( stderr, "\n" );
		*/
		
		ZeroMemory( &hid->overlapped, sizeof(OVERLAPPED) );
		
		_inner_hid_queue_read( hid );
		
		return;
	}
	
	free( hid->p_device_interface );
	
	free( DeviceDetail );
	destroyDeviceInfoList( hid->device_info );
}

